#import "CHFormTextListener.h"

#import <AddressBook/AddressBook.h>

#include "nsUnicharUtils.h"
#include "nsIContent.h"
#include "nsIDOMElement.h"
#include "nsIDOMMouseEvent.h"
#include "nsEmbedAPI.h"
#include "nsIDOMDocument.h"
#include "nsIDOMDocumentEvent.h"
#include "nsIDOMEventTarget.h"
#include "nsIDOMNSHTMLElement.h"
#include "nsIDOMNSHTMLInputElement.h"
#include "nsIDOMHTMLInputElement.h"
#include "nsIDOMWindow.h"
#include "nsIScriptGlobalObject.h"
#include "nsIDocument.h"

@interface NSArray (CHArrayRemoveObjects)

- (NSArray *)arrayByRemovingObjectsInArray:(NSArray *)anArray;

@end

@implementation NSArray (CHArrayRemoveObjects)

- (NSArray *)arrayByRemovingObjectsInArray:(NSArray *)anArray
{
	NSMutableArray *temp = [self mutableCopy];
	
	[temp removeObjectsInArray:anArray];
	
	NSArray *result = [temp copy];
	
	[temp release];
	
	return [result autorelease];
}

@end

@interface ABPerson (ABPersonReadableProperties)

+ (NSArray *)readableProperties;
+ (NSArray *)emailProperties;

@end

@implementation ABPerson (ABPersonReadableProperties)

// These properties are searched only in the 'Me' record
+ (NSArray *)readableProperties
{
	return [[ABPerson properties] arrayByRemovingObjectsInArray:
												[NSArray arrayWithObjects:
																							kABUIDProperty,
																							kABCreationDateProperty,
																							kABModificationDateProperty,
																							nil]];
}

// These properties are searched in all other address book entries
+ (NSArray *)emailProperties
{
	return [NSArray arrayWithObjects:
									kABEmailProperty,
									nil];
}

@end

@interface NSString (NSStringCaseInsensitivePrefix)

- (BOOL)hasCaseInsensitivePrefix:(NSString *)prefix;

@end

@implementation NSString (NSStringCaseInsensitivePrefix)

- (BOOL)hasCaseInsensitivePrefix:(NSString *)prefix
{
	return [[self lowercaseString] hasPrefix:[prefix lowercaseString]];
}

@end

// Look in a property (and it's sub-properties, if available) for the string toMatch
// eg. typing in "pers" would expand to "person@company.com" if the property was kABEmailProperty
// Returns nil if no match exists
NSString *FirstMatchingPropertyForValue(NSString *toMatch, id value, ABPropertyType type)
{
	switch (type)
	{
		case kABStringProperty:
			{
				if ([value hasCaseInsensitivePrefix:toMatch]) return value;
			}
			break;
		case kABIntegerProperty:
		case kABRealProperty:
			{
				if ([[value stringValue] hasCaseInsensitivePrefix:toMatch]) return value;
			}
			break;
		case kABDateProperty:
			{
				// TODO - multiple date formats (dd/mm/yy mm/dd/yy etc.)
			}
			break;
		case kABArrayProperty:
			break;											// TODO
		case kABDictionaryProperty:
			{
				unsigned i;
				NSArray *objects = [value allValues];
				
				for (i=0; i<[objects count]; i++)
				{
					id subValue = [objects objectAtIndex:i];
					NSString *result = FirstMatchingPropertyForValue(toMatch, subValue, kABStringProperty);
					
					if (result)
						return result;
				}
			}
			break;
		case kABDataProperty:
			break;											// TODO
		case kABMultiStringProperty:
		case kABMultiIntegerProperty:
		case kABMultiRealProperty:
		case kABMultiDateProperty:
		case kABMultiArrayProperty:
		case kABMultiDictionaryProperty:
		case kABMultiDataProperty:
			{
				unsigned i;
				ABPropertyType subType = (ABPropertyType)(type & (~kABMultiValueMask));
				
				for (i=0; i<[value count]; i++)
				{
					id subValue = [value valueAtIndex:i];
					NSString *result = FirstMatchingPropertyForValue(toMatch, subValue, subType);
					
					if (result)
						return result;
				}
			}
			break;
		case kABErrorInProperty:
			break;
	}
	
	return nil;
}

// Return the first property of person that matches toMatch
// Returns nil if no matches exist
NSString *FirstMatchingPropertyForPerson(NSString *toMatch, ABPerson *person, NSArray *properties)
{
	unsigned p;
	
	for (p=0; p<[properties count]; p++)
	{
		NSString *property = [properties objectAtIndex:p];
		id value = [person valueForProperty:property];
		ABPropertyType type = [ABPerson typeOfProperty:property];
		
		NSString *returnValue = FirstMatchingPropertyForValue(toMatch, value, type);
		
		if (returnValue != nil)
		{
			// Print full name & not just first name (user can always delete the last name part if they want)
			if ([property isEqual:kABFirstNameProperty])
			{
				returnValue = [NSString stringWithFormat:@"%@ %@", 
																	[person valueForProperty:kABFirstNameProperty],
																	[person valueForProperty:kABLastNameProperty]];
			}
			
			return returnValue;
		}
	}
	
	return nil;
}

// Complete the string by looking at values in the users Address Book
NSString *AutocompleteString(NSString *input)
{
	// First look for properties of the user
	NSString *result = FirstMatchingPropertyForPerson( input, 
																			[[ABAddressBook sharedAddressBook] me],
																			[ABPerson readableProperties]);
	
	// If that failed, look for properties that match in their oter contacts
	if (result == nil)
	{
		unsigned i;
		NSArray *people = [[ABAddressBook sharedAddressBook] people];
		
		for (i=0; i<[people count]; i++)
		{
			ABPerson *person = [people objectAtIndex:i];
			
			// Don't search -me twice
			if (![person isEqual:[[ABAddressBook sharedAddressBook] me]])
			{
				// Only search in -emailProperties, as searching every property may take too long (?)
				result = FirstMatchingPropertyForPerson(input, person, [ABPerson emailProperties]);
				
				if (result != nil)
				{
					break;
				}
			}
		}
	}
	
	return result;
}



NS_IMPL_ISUPPORTS2(CHFormTextListener, nsIDOMFormListener, nsIDOMEventListener);

CHFormTextListener::CHFormTextListener()
{
  NS_INIT_ISUPPORTS();
	
	lastInputLength = 0;
	lastTarget = nsnull;
	lastSearchEmpty = PR_FALSE;
}

CHFormTextListener::~CHFormTextListener()
{
}

NS_IMETHODIMP
CHFormTextListener::HandleEvent(nsIDOMEvent* aEvent) { return NS_OK; }

NS_IMETHODIMP
CHFormTextListener::Submit(nsIDOMEvent* aEvent) { return NS_OK; }

NS_IMETHODIMP
CHFormTextListener::Reset(nsIDOMEvent* aEvent) { return NS_OK; }

NS_IMETHODIMP
CHFormTextListener::Change(nsIDOMEvent* aEvent) { return NS_OK; }

NS_IMETHODIMP
CHFormTextListener::Select(nsIDOMEvent* aEvent) { return NS_OK; }

NS_IMETHODIMP
CHFormTextListener::Input(nsIDOMEvent* aEvent)
{
	nsCOMPtr<nsIDOMEventTarget> target;
  aEvent->GetTarget(getter_AddRefs(target));
	
  if (!target)
    return NS_OK;
	
  nsCOMPtr<nsIDOMHTMLInputElement> inputElement(do_QueryInterface(target));
	
  if (inputElement) 
	{
		nsAutoString type;
		inputElement->GetType(type);
		
		PRBool isPassword = type.Equals(NS_LITERAL_STRING("password"), nsCaseInsensitiveStringComparator());
		
		// Don't complete password fields
		if (!isPassword)
		{
			nsAutoString value;
			inputElement->GetValue(value);
			
			NSString *searchValue = [NSString stringWith_nsAString:value];
			
#if DEBUG
			NSLog(@"Attempting to autocomplete '%@'", searchValue);
#endif
			
			// If this is a different field in the form from the last autocompleted one
			// OR
			// (the user has typed in more text AND the last autocomplete was not empty)
			
			// Testing for !lastSearchEmpty means that once you have typed in n characters that do not autocomplete to anything in your address book, then there isn't any point in searching until the user either deletes some characters or moves to a different text field
			if ((lastTarget != inputElement) || ((value.Length() > lastInputLength) && !lastSearchEmpty))
			{
				nsCOMPtr<nsIDOMNSHTMLInputElement> inputExtra(do_QueryInterface(target));
				
				if (inputExtra) 
				{
					PRInt32 selStart, selEnd;
					
					inputExtra->GetSelectionStart(&selStart);
					inputExtra->GetSelectionEnd(&selEnd);
					
					// If the user is typing at the end of the text field
					if ((selStart == selEnd) && (selEnd == (PRInt32)value.Length()))
					{
						NSString *result = AutocompleteString(searchValue);
#if DEBUG
						NSLog(@"    Got autocomplete result '%@'", result);
#endif
						if (result)
						{
							// Append rest of match to end of text field and select it
							lastSearchEmpty = PR_FALSE;
							
							NSString *toAppend = [result substringWithRange:NSMakeRange(selEnd, [result length] - value.Length())];
							
							PRUnichar* buffer = [toAppend createNewUnicodeBuffer];
							
							nsAutoString newValue(value);
							newValue.Append(buffer);
							
							inputElement->SetValue(newValue);
							
							inputExtra->SetSelectionRange(selEnd, newValue.Length());
							
							nsMemory::Free(buffer);
						}
						else
						{
							lastSearchEmpty = PR_TRUE;
						}
					}
				}
			}
			else
			{
#if DEBUG
				NSLog(@"    Not searching for matches - not typing at end of selection or no matches found last time");
#endif
				lastSearchEmpty = PR_FALSE;
			}
			
			lastInputLength = value.Length();
			lastTarget = inputElement;
		}
		else
		{
#if DEBUG
			NSLog(@"    Will not autocomplete password field");
#endif
		}
	}
	
  return NS_OK;
}

